#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
# Introduction

This documentation describes the cobwood package to analyse global forest
products markets. The package is distributed on the python pacage index at
https://pypi.org/project/cobwood/. The source code is available at
https://bioeconomy.gitlab.io/cobwood/cobwood/

Key Features:

- **Panel Data Structure**: The package represents international forest products market
  data through 2 dimensional arrays with multiple years
  and countries, enabling time-series and cross-sectional analysis.

- **Data Handling with Xarray**: Utilizes Xarray datasets to efficiently manipulate
  multi-dimensional data structures. X array is tightly integrated with pandas.
  Conversion to and from pandas data frames is very straightforward. Xarray datasets are
  saved on disk using the NetCDF format. This format has the advantage of providing good
  metadata descriptors. NetCDF is a standard data format used in earth systems
  modelling, that will just help this model become a component of a greater modelling
  tool chain.


# Models

The goal of cobwood is to provide a readable and reusable data structure to implement many forest sector models.


## GFPMx

Currently, only the GFFPMx model is available. We will now illustrate how to prepare the input data for that model, and how to define a scenario. Not that many scenarios can be defined from the same input data, by changing some of the variables, such as the GDP projections. We will then explain how to run the model.

0. The input data can be downloaded from the website of the university of
  Wisconsin and converted to CSV files. You need to do this step only once and
  can ignore it later on. Convert data to CSV files inside the cobwood_data
  directory:

    >>> from cobwood.gfpmx_spreadsheet_to_csv import gfpmx_spreadsheet_to_csv
    >>> gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-base2021.xlsb")

1. In the cobwood_data directory, write the following configuration files and call it `scenario/base_2021.yaml`

```
input_dir: "gfpmx_base2021"
base_year: 2021
description: "Reproduce the GFPMX base 2021 scenario"
```

2. Load the input data into a [GFPMX](cobwood/gfpmx.html) model object.

    >>> from cobwood.gfpmx import GFPMX
    >>> gfpmxb2021 = GFPMX(scenario="base_2021", rerun=True)

3. Run the model.At each step compare with the reference model run inside the
Excel Sheet:

    >>> gfpmxb2021.run(compare=True, strict=False)

4. Explore the model output tables and make plots.

    >>> print(gfpmxb2021["sawn"])
    >>> print(gfpmxb2021["sawn"]["cons"])
    >>> gfpmxb2021.facet_plot_by_var("indround")

"""

# Build in modules
from pathlib import Path
import os

# Where is the data, default case #
data_dir = Path("~/repos/cobwood_data/")
data_dir = data_dir.expanduser()

# TODO: remove when it is replaced everywhere by cobwood.data_dir. Maybe it's
# better to keep it that way explicitly in case we import data_dir from another
# package?
cobwood_data_dir = data_dir

# But you can override that with an environment variable #
if os.environ.get("COBWOOD_DATA"):
    data_dir = Path(os.environ["COBWOOD_DATA"])


def create_data_dir(path: str):
    """Create a sub directory of `data_dir`

    Example:

        >>> import cobwood
        >>> test_sub_dir = cobwood.create_data_dir("test")
        >>> print(test_sub_dir)

    """
    sub_dir = data_dir / path
    if not sub_dir.exists():
        sub_dir.mkdir()
    return sub_dir



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Written by Paul Rougieux.

JRC biomass Project.
Unit D1 Bioeconomy.

Copied from https://docs.python.org/3/howto/logging-cookbook.html

Create a cobwood logger with file handler:

    >>> import cobwood.logger

This can be then used directly from any sub module:

    >>> import logging
    >>> logger = logging.getLogger('cobwood.sub_module')
    >>> logger.info("Doing this and that")

"""

# Third party modules
import logging

# Internal modules
from cobwood import cobwood_data_dir

# create logger with 'cobwood'
logger = logging.getLogger("cobwood")
logger.setLevel(logging.DEBUG)
# create file handler which logs even debug messages
fh = logging.FileHandler(cobwood_data_dir / "gfpmx.log")
fh.setLevel(logging.DEBUG)
# create console handler with a higher log level
ch = logging.StreamHandler()
ch.setLevel(logging.ERROR)
# create formatter and add it to the handlers
formatter = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")
fh.setFormatter(formatter)
ch.setFormatter(formatter)
# add the handlers to the logger
logger.addHandler(fh)
logger.addHandler(ch)

logger.info(
    "Created a logger with file handler %s.", str(cobwood_data_dir / "gfpmx.log")
)
# Usage
# from cobwood.gfpmx_runner import gfpmx_runner
# logger.info('Calling gfpmx_runner.run_next_step().')
# gfpmx_runner.run_next_step()
# logger.info('Finished gfpmx_runner.run_next_step().')



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
""" This script is deprecated, see the xarray version.

Functions to implement equations in the paper

    Buongiorno, J. (2021). GFPMX: A Cobweb Model of the Global Forest Sector, with
    an Application to the Impact of the COVID-19 Pandemic. Sustainability, 13(10),
    5507.

    https://www.mdpi.com/2071-1050/13/10/5507

    Local link

    ~/repos//bioeconomy_papers/literature/buongiorno2021_gfpmx_a_cobweb_model_of_the_global_forest_sector_with_an_application_to_the_impact_of_the_covid_19_pandemic.pdf

Excel file "~/large_models/GFPMX-8-6-2021.xlsx"

A script to compute GFPMX equations recursively in a time loop

TODO: rename this script to compute_all_equations_with_pandas.py

Run this file with:

     ipython -i ~/repos/cobwood/scripts/compute_all_equations.py

Equation numbers in this script refer to the paper:

    Buongiorno, J. (2021). GFPMX: A Cobweb Model of the Global Forest Sector, with
    an Application to the Impact of the COVID-19 Pandemic. Sustainability, 13(10),
    5507.

    https://www.mdpi.com/2071-1050/13/10/5507

    Local link

    ~/repos//bioeconomy_papers/literature/buongiorno2021_gfpmx_a_cobweb_model_of_the_global_forest_sector_with_an_application_to_the_impact_of_the_covid_19_pandemic.pdf

Excel file "~/large_models/GFPMX-8-6-2021.xlsx"
"""
import numpy

from cobwood.gfpmx_data import GFPMXData

# These functions have been moved back out of the package, into this script
# from cobwood.gfpmx_functions_pandas import (
#     shift_index,
#     compute_demand,
#     compute_import_demand,
#     compute_export_supply,
#     compute_domestic_production,
#     compute_world_price,
#     compute_local_price,
# )
from cobwood.gfpmx_qaqc import (
    check_world_aggregates,
    check_nrows_years_countries,
    compare_to_original_cobwood,
)

# Load data
gfpmx_data = GFPMXData(data_dir="gfpmx_8_6_2021")
indround_agg = gfpmx_data.get_agg_rows("indround")
pulp_agg = gfpmx_data.get_agg_rows("pulp")
pulp = gfpmx_data.get_country_rows("pulp", ["gdp"])
sawn = gfpmx_data.get_country_rows("sawn", ["gdp"])
sawn_agg = gfpmx_data.get_agg_rows("sawn", ["gdp"])
fuel = gfpmx_data.get_country_rows("fuel", ["gdp"])
fuel_agg = gfpmx_data.get_agg_rows("fuel", ["gdp"])
# TODO: remove rows containing "World prod/cons" or simply that don't have a FAOSTAT name from
# the input data pre-processed in "~/rp/cobwood/scripts/gfpmx_spreadsheet_to_csv.py"
panel = gfpmx_data.get_country_rows("panel", ["gdp"])
panel_agg = gfpmx_data.get_agg_rows("panel", ["gdp"])
paper = gfpmx_data.get_country_rows("paper", ["gdp"])
paper_agg = gfpmx_data.get_agg_rows("paper", ["gdp"])

# Check that world aggregates correspond to the sum of countries
check_world_aggregates(fuel, fuel_agg)
check_world_aggregates(sawn, sawn_agg)
check_world_aggregates(panel, panel_agg, rtol=1e-4)
check_world_aggregates(paper, paper_agg)

# Display the number of countries and years
print(check_nrows_years_countries(fuel, "fuel"))
print(check_nrows_years_countries(sawn, "sawn"))
print(check_nrows_years_countries(panel, "panel"))
print(check_nrows_years_countries(paper, "paper"))


def shift_index(x):
    """Update the index of a lagged variable
    To store last year's prices in a "price_lag" column at year t.
    We first need to update the index from year t-1 to year t.
    Otherwise assignation of .loc[[t], "price_lag"] would contain NA values.

    :param x pandas series input variable indexed by year and country
    :return pandas series
    """
    df = x.reset_index()
    df["year"] = df["year"] + 1
    x_lag = df.set_index(x.index.names)[x.name]
    return x_lag


def compute_demand(df):
    """GFPMX demand equation 1"""
    return (
        df["cons_constant"]
        * df["price_lag"].pow(df["cons_price_elasticity"])
        * df["gdp"].pow(df["cons_gdp_elasticity"])
    )


def compute_import_demand(df):
    """GFPMX import demand equation 4"""
    return (
        df["imp_constant"]
        * (df["price_lag"] * (1 + df["tariff_lag"])).pow(df["imp_price_elasticity"])
        * df["gdp"].pow(df["imp_gdp_elasticity"])
    )


def compute_export_supply(df):
    """GFPMX export supply equation 7"""
    world_imp = df["imp"].sum()
    exp = df["exp_marginal_propensity_to_export"] * world_imp + df["exp_constant"]
    # Use numpy.maximum to propagate NA values
    return numpy.maximum(exp, 0)


def compute_domestic_production(df):
    """GFPMX domestic production equation 8

    Replace negative values by zero"""
    prod = df["cons"] + df["exp"] - df["imp"]
    prod.loc[prod < 0] = 0
    return prod


def compute_world_price(s_world, s_primary_world):
    """GFPMX world price as a function of the input price
    - World price of f, s, u, l as a function of the roundwood price equation 10
    - World price of paper as a function of the pulp price equation 11
    """
    return s_world["price_constant"] * pow(
        s_primary_world["price"], s_world["price_input_elast"]
    )


def compute_local_price(df, s_world):
    """GFPMX local price as a function of the world price equation 12"""
    # world_price = s_world.xs("WORLD", level="country")["price"].iat[0]
    price = df["price_constant"] * pow(
        s_world.loc["price2"], df["price_world_price_elasticity"]
    )
    return price


def compute_end_product_time_step(t, df, df_agg, df_prim_agg):
    """Compute a time step for the given product

    Attention! This will edit the input data frame by reference.
    It will edit the following new columns:

        ['price_lag', 'tariff_lag', 'cons2', 'imp2',
        'exp2', 'prod2', 'price2']

    :param df data frame: end product data at time step t
    :param df_agg data frame: end product data aggregated
    :param df data frame: end product data at time step t
    :return computed variables at time step t

    Usage:

        >>> compute_end_product_time_step(2019, sawn, sawn_agg, indround_agg)
    """
    # Keep `[t]` in square braquets so that years is kept in the index on both sides
    df.loc[[t], "price_lag"] = shift_index(df.loc[[t - 1], "price"])
    df.loc[[t], "tariff_lag"] = shift_index(df.loc[[t - 1], "tariff"])
    df.loc[[t], "cons2"] = compute_demand(df.loc[[t]])
    df.loc[[t], "imp2"] = compute_import_demand(df.loc[[t]])
    df.loc[[t], "exp2"] = compute_export_supply(df.loc[[t]])
    df.loc[[t], "prod2"] = compute_domestic_production(df.loc[[t]])
    df_agg.loc[(t, "WORLD"), "price2"] = compute_world_price(
        df_agg.loc[(t, "WORLD")], df_prim_agg.loc[(t, "WORLD")]
    )
    df.loc[[t], "price2"] = compute_local_price(df.loc[[t]], df_agg.loc[(t, "WORLD")])


years = sawn.index.to_frame()["year"].unique()
# Start one year after the base year so price_{t-1} exists already
for t in range(gfpmx_data.base_year + 1, years.max() + 1):
    # The world price of secondary products are based in the price of industrial roundwood
    compute_end_product_time_step(t, fuel, fuel_agg, indround_agg)
    compute_end_product_time_step(t, sawn, sawn_agg, indround_agg)
    compute_end_product_time_step(t, panel, panel_agg, indround_agg)
    # The world price of paper and paper board is based on the price of wood pulp
    compute_end_product_time_step(t, paper, paper_agg, pulp_agg)
    # Compute domestic demand for wood pulp


compare_to_original_cobwood(fuel)
compare_to_original_cobwood(sawn)
compare_to_original_cobwood(panel)
compare_to_original_cobwood(paper)



import yaml
from pathlib import Path
from typing import Union, Dict, Any


def parse_scenario_yaml(yaml_path: Union[str, Path]) -> Dict[str, Any]:
    """
    Parse the YAML configuration file for the given scenario.

    Parameters:
    -----------
    yaml_path : Union[str, Path]
        Path to the scenario YAML file. Can be a string or pathlib.Path object.
        If a string is provided, it will be converted to a Path object.

    Returns:
    --------
    Dict[str, Any]
        Dictionary containing the configuration parameters

    Raises:
    -------
    FileNotFoundError
        If the specified YAML file does not exist
    ValueError
        If the YAML file cannot be parsed or is missing required fields

    Examples:
    ---------
    Parse a scenario configuration file:

        >>> from cobwood.scenario import parse_scenario_yaml
        >>> from pathlib import Path
        >>> config = parse_scenario_yaml(Path.home() / "repos/cobwood_data/scenario/pikfair_fel1.yaml")

    Required fields in the YAML file:
    - input_dir: Directory containing input data
    - base_year: Base year for the scenario
    """
    # Convert string to Path object if necessary
    if not isinstance(yaml_path, Path):
        yaml_path = Path(yaml_path)

    # Check if file exists
    if not yaml_path.exists():
        msg = f"Configuration file not found at: {yaml_path}"
        raise FileNotFoundError(msg)

    # Parse YAML file
    with open(yaml_path, "r") as yaml_file:
        try:
            config = yaml.safe_load(yaml_file)
        except yaml.YAMLError as e:
            msg = f"Error parsing YAML configuration file {yaml_path}: {e}"
            raise ValueError(msg) from e

    # Validate required fields
    required_fields = ["input_dir", "base_year"]
    for field in required_fields:
        if field not in config:
            msg = f"Missing required field '{field}' in configuration file: {yaml_path}"
            raise ValueError(msg)

    return config



""" Run tests with

    cd ~/repos/cobwood/
    pytest

"""

# Pylint exception needed https://github.com/pylint-dev/pylint/issues/6531
# pylint: disable=redefined-outer-name
import pytest
import xarray
from cobwood.gfpmx_equations import (
    consumption,
    consumption_pulp,
    consumption_indround,
    import_demand,
    import_demand_pulp,
    # import_demand_indround,
    # export_supply   ,
    # production      ,
    # world_price     ,
    # world_price_indround   ,
    # local_price     ,
    # forest_stock    ,
)


@pytest.fixture
def primary_product_dataset():
    """Create a sample dataset for testing"""
    ds = xarray.Dataset(
        {
            "cons_constant": xarray.DataArray([2, 3, 4], dims=["country"]),
            "imp_constant": xarray.DataArray([2, 3, 4], dims=["country"]),
            "price": xarray.DataArray(
                [[1, 2], [3, 4], [5, 6]], dims=["country", "year"]
            ),
            "cons_paper_production_elasticity": xarray.DataArray(
                [0.9, 1.0, 0.8], dims=["country"]
            ),
            "cons_price_elasticity": xarray.DataArray(
                [0.5, 0.6, 0.7], dims=["country"]
            ),
            "cons_products_elasticity": xarray.DataArray(
                [0.5, 0.6, 0.7], dims=["country"]
            ),
            "tariff": xarray.DataArray(
                [[0.1, 0.2], [0.3, 0.4], [0.5, 0.6]], dims=["country", "year"]
            ),
            "c": xarray.DataArray([True, True, True], dims=["country"]),
        }
    )
    return ds


@pytest.fixture
def secondary_product_dataset():
    """Create a sample dataset for testing"""
    ds = xarray.Dataset(
        {
            "cons_constant": xarray.DataArray([2, 3, 4], dims=["country"]),
            "imp_constant": xarray.DataArray([2, 3, 4], dims=["country"]),
            "price": xarray.DataArray(
                [[1, 2], [3, 4], [5, 6]], dims=["country", "year"]
            ),
            "gdp": xarray.DataArray(
                [[100, 200], [300, 400], [500, 600]], dims=["country", "year"]
            ),
            "prod": xarray.DataArray(
                [[100, 200], [300, 400], [500, 600]], dims=["country", "year"]
            ),
            "cons_price_elasticity": xarray.DataArray(
                [0.5, 0.6, 0.7], dims=["country"]
            ),
            "imp_price_elasticity": xarray.DataArray([0.5, 0.6, 0.7], dims=["country"]),
            "cons_gdp_elasticity": xarray.DataArray([0.8, 0.9, 1.0], dims=["country"]),
            "tariff": xarray.DataArray(
                [[0.1, 0.2], [0.3, 0.4], [0.5, 0.6]], dims=["country", "year"]
            ),
            "c": xarray.DataArray([True, True, True], dims=["country"]),
        }
    )
    return ds


def test_consumption(secondary_product_dataset):
    """Test the consumption function"""
    ds = secondary_product_dataset
    t = 1
    expected_result = xarray.DataArray(
        [138.62896863, 1274.23051055, 7404.40635264], dims=["country"]
    )
    result = consumption(ds, t)
    xarray.testing.assert_allclose(result, expected_result)


def test_consumption_pulp(primary_product_dataset, secondary_product_dataset):
    """Test the consumption_pulp function"""
    ds = primary_product_dataset
    ds_paper = secondary_product_dataset
    t = 1
    expected_result = xarray.DataArray(
        [235.48160746, 2319.81845392, 2059.96572644], dims=["country"]
    )
    result = consumption_pulp(ds, ds_paper, t)
    xarray.testing.assert_allclose(result, expected_result)


def test_consumption_indround(primary_product_dataset, secondary_product_dataset):
    """Test the consumption_indround function"""
    ds = primary_product_dataset
    ds["cons_constant"].loc[2] = -10
    ds_sawn = secondary_product_dataset
    ds_panel = secondary_product_dataset
    ds_pulp = secondary_product_dataset
    t = 1
    compatible_mode = False
    expected_result = xarray.DataArray(
        [48.98979486, 408.22796795, 0],
        dims=["country"],
    )
    result = consumption_indround(ds, ds_sawn, ds_panel, ds_pulp, t, compatible_mode)
    xarray.testing.assert_allclose(result, expected_result)


def test_consumption_indround_compatible_mode(
    primary_product_dataset, secondary_product_dataset
):
    """Test the consumption_indround function with compatible mode"""
    ds = primary_product_dataset
    ds["cons_constant"].loc[2] = -10
    ds_sawn = secondary_product_dataset
    ds_panel = secondary_product_dataset
    ds_pulp = secondary_product_dataset
    t = 1
    compatible_mode = True
    expected_result = xarray.DataArray(
        [48.98979486, 408.22796795, -5860.973485],
        dims=["country"],
    )
    result = consumption_indround(ds, ds_sawn, ds_panel, ds_pulp, t, compatible_mode)
    xarray.testing.assert_allclose(result, expected_result)


def test_import_demand(secondary_product_dataset):
    """Test the import_demand function"""
    ds = secondary_product_dataset
    t = 1
    expected_result = xarray.DataArray(
        [1.043081, 2.592547, 4.999707],
        dims=["country"],
    )
    result = import_demand(ds, t)
    xarray.testing.assert_allclose(result, expected_result)


def test_import_demand_pulp(primary_product_dataset, secondary_product_dataset):
    """Test the import_demand_pulp function"""
    ds = primary_product_dataset
    ds_paper = secondary_product_dataset
    t = 1

    expected_result = xarray.DataArray(
        [1.043081, 2.592547, 2.869484],
        dims=["country"],
    )
    result = import_demand_pulp(ds, ds_paper, t)
    xarray.testing.assert_allclose(result, expected_result)



""" Extract data to compute the Harvested Wood Products Sink

- Load GFPMx output for the SSP2 and Fair pathways

- Rescale the harvest expected to the harvest actually provided and distinguish
  coniferous from Broadleaves. Use the CBM output table of harvest expected
  provided aggregated by coniferous and broadleaves.

Usage:

    ipython -i ~/repos/cobwood/scripts/extract_data_for_hwp_sink_computations.py

"""

import sys
import warnings
import numpy as np
from cobwood.gfpmx import GFPMX
from biotrade.faostat import faostat
import cobwood

# Load CBM output data
eu_cbm_explore_path = cobwood.data_dir.parent / "eu_cbm" / "eu_cbm_explore"
scenario_dir = eu_cbm_explore_path / "scenarios" / "ssp2_fair_degrowth"
# Load data and plotting functions from the prepare script in the same directory
sys.path.append(str(scenario_dir))
from ssp2_fair_owc_prepare_data import hexprovft_wide  # noqa: E402
from ssp2_fair_owc_prepare_data import hexprov_wide  # noqa: E402


hwp_dir = cobwood.data_dir / "gfpmx_output" / "hwp"
hwp_dir.mkdir(exist_ok=True)

eu_countries = faostat.country_groups.eu_country_names
eu_countries += ["Netherlands"]

# Load output data, after a run has already been completed
gfpmx_pikssp2 = GFPMX(
    input_dir="gfpmx_base2021", base_year=2021, scenario="pikssp2_fel1"
)
gfpmx_pikfair = GFPMX(
    input_dir="gfpmx_base2021", base_year=2021, scenario="pikfair_fel1"
)

SELECTED_VARIABLES = [
    "cons",
    "prod",
    "exp",
    "imp",
]
PRODUCTS = ["indround", "fuel", "sawn", "panel", "pulp", "paper"]
print("Check units")
for product in PRODUCTS:
    for var in SELECTED_VARIABLES:
        print(product, var, gfpmx_pikssp2[product][var].unit)
msg = "There is an issue with the unit, it should be in tons instead of m3.\n"
msg += "see https://gitlab.com/bioeconomy/cobwood/cobwood/-/issues/11"
warnings.warn(msg)


def get_gfpmx_df(gfpmx_scenario, product):
    """Extract GFPMX projection output to a data frame"""
    ds = gfpmx_scenario[product]
    df = (
        ds[SELECTED_VARIABLES]
        .to_dataframe()
        .reset_index()
        .sort_values(["country", "year"])
    )
    df["pathway"] = gfpmx_scenario.scenario
    df["product"] = ds.product
    if product in ["paper", "pulp"]:
        df["unit"] = "1000 t"
    else:
        df["unit"] = "1000 m3"
    selector = df["country"].isin(eu_countries)
    # Place last columns first
    cols = list(df.columns)
    cols = cols[-3:] + cols[:-3]
    df = df.loc[selector].copy()[cols]
    return df


# For example
indround_fair = get_gfpmx_df(gfpmx_pikfair, "indround")

###########
# Rescale #
###########
# Aggregate harvest by combo_name, country, year and con_broad
index = ["combo_name", "iso2_code", "country", "year"]
cols = [
    "irw_need",
    "fw_colat",
    "fw_need",
    "amount_exp_hat",
    "harvest_exp_hat",
    "area",
    "to_product",
    "harvest_prov_ub",
    "harvest_prov_ob",
]
hexprov_cb = (
    hexprovft_wide.groupby(index + ["con_broad"])[cols].agg("sum").reset_index()
)
hexprov_cb["con_broad"] = "harvest_prov_ub_" + hexprov_cb["con_broad"]
# Reshape to wider format to get harvest_prov_ub_broad and harvest_prov_ub_con
hexprov_cb = hexprov_cb.pivot(
    columns="con_broad", index=index, values="harvest_prov_ub"
).reset_index()

# Convert data from the economic model from 1000 m3 ub to m3 ub
vars_to_convert = ["rw_demand", "fw_demand", "irw_demand"]
hexprov_wide[vars_to_convert] *= 1e3

# Merge annual harvest with harvest by con and broad
hexprov_cb_2 = hexprov_cb.merge(hexprov_wide, on=index)

# Check irw demand is equal to irw_need + irw_predetermined
locator = hexprov_wide["irw_need"] > 0
np.testing.assert_allclose(
    hexprov_wide.loc[locator, ["irw_need", "irw_predetermined"]].sum(axis=1),
    hexprov_wide.loc[locator, "irw_demand"],
    rtol=0.01,
)

# Check fw demand is equal to fw_colat + fw_need + fw_predetermined
locator = hexprov_wide["fw_need"] > 0
np.testing.assert_allclose(
    hexprov_wide.loc[locator, ["fw_colat", "fw_need", "fw_predetermined"]].sum(axis=1),
    hexprov_wide.loc[locator, "fw_demand"],
    rtol=0.01,
)


#############################
# Write output to CSV files #
#############################
def extract_to_csv(gfpmx_scenario, product):
    """Extract GFPMx projection to a csv file"""
    df = get_gfpmx_df(gfpmx_scenario=gfpmx_scenario, product=product)
    df.to_csv(hwp_dir / f"{gfpmx_scenario.scenario}_{product}_eu.csv", index=False)


# for product in PRODUCTS:
#     extract_to_csv(gfpmx_pikfair, product)
#     extract_to_csv(gfpmx_pikssp2, product)



"""Run the GFPMX model

Usage:

    >>> ipython -i ~/repos/cobwood/scripts/run_gfpmx.py

"""
# from cobwood.gfpmx_spreadsheet_to_csv import gfpmx_spreadsheet_to_csv
# from cobwood.gfpmx_data import GFPMXData
from cobwood.gfpmx import GFPMX

########################################
# Convert the Excel spreadsheet to CSV #
########################################
# gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-8-6-2021.xlsx")
# gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-base2020.xlsx")
# gfpmx_spreadsheet_to_csv("~/large_models/GFPMX-base2021.xlsb")

###################################################
# Load data only to debug dataset creation issues #
###################################################
# gfpmx_data_b2018 = GFPMXData(data_dir="gfpmx_8_6_2021")
# gfpmx_data_b2020 = GFPMXData(data_dir="gfpmx_base2020")
# gfpmx_data_b2021 = GFPMXData(data_dir="gfpmx_base2021")

#############################
# Instantiate model objects #
#############################
# gfpmxb2018 = GFPMX(data_dir="gfpmx_8_6_2021", base_year=2018)
# gfpmxb2020 = GFPMX(data_dir="gfpmx_base2020", base_year=2020)
gfpmxb2021 = GFPMX(
    input_dir="gfpmx_base2021", base_year=2021, scenario="base_2021", rerun=True
)

#######
# Run #
#######
gfpmxb2021.run(compare=True, strict=False)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
A script to compute GFPMX sawnwood equations recursively in a time loop

Run this file with:

     ipython -i ~/repos/cobwood/scripts/compute_sawnwood_equations.py

Equation numbers in this script refer to the paper:

    Buongiorno, J. (2021). GFPMX: A Cobweb Model of the Global Forest Sector, with
    an Application to the Impact of the COVID-19 Pandemic. Sustainability, 13(10),
    5507.

    https://www.mdpi.com/2071-1050/13/10/5507

    Local link

    ~/repos//bioeconomy_papers/literature/buongiorno2021_gfpmx_a_cobweb_model_of_the_global_forest_sector_with_an_application_to_the_impact_of_the_covid_19_pandemic.pdf

Excel file "~/large_models/GFPMX-8-6-2021.xlsx"
"""

# Third party modules
from numpy.testing import assert_allclose

# Internal modules
from cobwood.gfpmx_data import gfpmx_data
from cobwood.gfpmx_functions import (
    shift_index,
    compute_demand,
    compute_import_demand,
    compute_export_supply,
    compute_domestic_production,
    compute_world_price,
    compute_local_price,
)

# Load sawnwood data
sawn = gfpmx_data.get_country_rows("sawn", ["gdp"])
sawn_agg = gfpmx_data.get_agg_rows("sawn", ["gdp"])
# Load industrial roundwood aggregates
indround_agg = gfpmx_data.get_agg_rows("indround")

# Check that world aggregates correspond to the sum of countries
gfpmx_data.check_world_aggregates("sawn", ["gdp"])


# Display the number of years and number of countries
years = sawn.index.to_frame()["year"].unique()
countries = sawn.index.to_frame()["country"].unique()
print("Years:", years)
print("Countries:", countries)
print("Number of lines in the sawn data frame:", len(sawn))
print(
    "Number of years time the number of countries: ",
    len(years),
    "*",
    len(countries),
    "=",
    len(years) * len(countries),
)

# Start one year after the base year so price_{t-1} exists already
for t in range(gfpmx_data.base_year + 1, years.max() + 1):
    # Keep `[t]` in square braquets so that years is kept in the index on both sides
    sawn.loc[[t], "price_lag"] = shift_index(sawn.loc[[t - 1], "price"])
    sawn.loc[[t], "tariff_lag"] = shift_index(sawn.loc[[t - 1], "tariff"])
    sawn.loc[[t], "cons2"] = compute_demand(sawn.loc[[t]])
    sawn.loc[[t], "imp2"] = compute_import_demand(sawn.loc[[t]])
    sawn.loc[[t], "exp2"] = compute_export_supply(sawn.loc[[t]])
    sawn.loc[[t], "prod2"] = compute_domestic_production(sawn.loc[[t]])
    sawn_agg.loc[(t, "WORLD"), "price2"] = compute_world_price(
        sawn_agg.loc[(t, "WORLD")], indround_agg.loc[(t, "WORLD")]
    )
    sawn.loc[[t], "price2"] = compute_local_price(
        sawn.loc[[t]], sawn_agg.loc[(t, "WORLD")]
    )


##################################
# Post processing quality checks #
##################################
# TODO use np.testing.assert_allclose()
# Check that the computed values correspond to the original GFPMx values
# Compare only values after the base year
sawn_comp = sawn.query("year > @gfpmx_data.base_year + 1")
for var in ["cons", "imp", "exp", "prod", "price"]:
    try:
        assert_allclose(sawn_comp[var + "2"], sawn_comp[var], rtol=1e-6)
    except AssertionError as e:
        print("The", var, "variable does not match with original GFPMx data:\n", str(e))


# sawn.to_csv("/tmp/sawn.csv") # Open with gx
